package convertquery

import (
	"encoding/json"
	"fmt"
)

// Service is a model for the convertquery use case
type Service struct {
}

// NewService creates a new convertquery service
func NewService() *Service {
	return &Service{}
}

// Constraint datatypes
// text		  MatchTypes: exact/contains/startswith/endswith
// number   MatchTypes: GT/LT/EQ
// bool     MatchTypes: EQ/NEQ

// Ranges dus tussen 10 half 5 bijv.

// Struct used for JSON conversion
type parsedJSON struct {
	Return    returnStruct     //`json:"return"`
	Entities  []entityStruct   //`json:"entities"`
	Relations []relationStruct //`json:"relations"`
}

type returnStruct struct {
	Entities  []int //`json:"entities"`
	Relations []int //`json:"relation"`
}

type entityStruct struct {
	Type        string             //`json:"type"`
	Constraints []constraintStruct //`json:"constraints"`
}
type relationStruct struct {
	Type        string //`json:"type"`
	EntityFrom  int
	EntityTo    int
	Depth       searchDepthStruct
	Constraints []constraintStruct
}
type searchDepthStruct struct {
	Min int
	Max int
}

type constraintStruct struct {
	Attribute string //`json:"attribute"`
	Value     string //`json:"value"`
	DataType  string //`json:"dataType"`
	MatchType string //`json:"matchType"`
}

// ConvertQuery converts a json string to an AQL query
func (s *Service) ConvertQuery(jsonMsg *[]byte) (*string, error) {

	jsonStruct, err := convertJSONToStruct(jsonMsg)
	if err != nil {
		fmt.Println(err)
		return nil, err
	}

	result := createQuery(jsonStruct)
	return result, nil
}

func convertJSONToStruct(jsonMsg *[]byte) (*parsedJSON, error) {
	jsonStruct := parsedJSON{}
	err := json.Unmarshal(*jsonMsg, &jsonStruct)

	if err != nil {
		return nil, err
	}

	return &jsonStruct, nil
}

func createQuery(jsonQuery *parsedJSON) *string {
	// GROTE SIDENOTE:
	// Vrij zeker dat een query waar alléén edges worden opgevraagd (#4)
	// niet wordt gesupport door zowel de result parser als de frontend receiver

	var (
		relationsToReturn []string
		nodesToReturn     []string
		nodeUnion         string
		relationUnion     string
	)

	// Loop over all relations
	ret := ""

	for i, relation := range jsonQuery.Relations {
		relationName := fmt.Sprintf("r%v", i)

		if relation.EntityFrom != -1 {
			// if there is a from-node
			// create the let for this node
			fromName := fmt.Sprintf("n%v", relation.EntityFrom)
			ret += *createNodeLet(&jsonQuery.Entities[relation.EntityFrom], &fromName)

			ret += *createRelationLetWithFromEntity(&relation, relationName, &jsonQuery.Entities)
		} else if relation.EntityTo != -1 {
			// if there is only a to-node
			toName := fmt.Sprintf("n%v", relation.EntityTo)
			ret += *createNodeLet(&jsonQuery.Entities[relation.EntityTo], &toName)

			ret += *createRelationLetWithOnlyToEntity(&relation, relationName, &jsonQuery.Entities)
			// Add this relation to the list
		} else {
			fmt.Println("Relation-only queries are currently not supported")
			continue
		}

		// Add this relation to the list
		relationsToReturn = append(relationsToReturn, relationName)
	}

	// Add node let statements for nodes that are not yet returned
	// Create a set from all the entity-from's and entity-to's, to check if they are returned
	nodeSet := make(map[int]bool)
	for _, relation := range jsonQuery.Relations {
		nodeSet[relation.EntityFrom] = true
		nodeSet[relation.EntityTo] = true
	}

	// Check if the entities to return are already returned
	for _, entityIndex := range jsonQuery.Return.Entities {
		if !nodeSet[entityIndex] {
			// If not, return this node
			name := fmt.Sprintf("n%v", entityIndex)
			ret += *createNodeLet(&jsonQuery.Entities[entityIndex], &name)

			// Add this node to the list
			nodesToReturn = append(nodesToReturn, name)
		}
	}

	// Create UNION statements that create unique lists of all the nodes and relations
	nodeUnion = "\nLET nodes = first(RETURN UNION_DISTINCT("
	for _, relation := range relationsToReturn {
		nodeUnion += fmt.Sprintf("flatten(%v[**].vertices), ", relation)
	}

	for _, node := range nodesToReturn {
		nodeUnion += fmt.Sprintf("%v,", node)
	}
	nodeUnion += "[],[]))\n"

	relationUnion = "LET edges = first(RETURN UNION_DISTINCT("
	for _, relation := range relationsToReturn {
		relationUnion += fmt.Sprintf("flatten(%v[**].edges), ", relation)
	}
	relationUnion += "[],[]))\n"

	ret += nodeUnion + relationUnion
	ret += "RETURN {\"vertices\":nodes, \"edges\":edges }"

	return &ret
}

func createNodeLet(node *entityStruct, name *string) *string {
	header := fmt.Sprintf("LET %v = (\n\tFOR x IN %v \n", *name, node.Type)
	footer := "\tRETURN x\n)\n"
	constraints := *createConstraintStatements(&node.Constraints, "x", false)

	ret := header + constraints + footer
	return &ret
}

func createRelationLetWithFromEntity(relation *relationStruct, name string, entities *[]entityStruct) *string {
	header := fmt.Sprintf("LET %v = (\n\tFOR x IN n%v \n", name, relation.EntityFrom)
	forStatement := fmt.Sprintf("\tFOR v, e, p IN %v..%v OUTBOUND x %s \n", relation.Depth.Min, relation.Depth.Max, relation.Type)

	// Guarantees that there is no path returned with a duplicate edge
	// This way there are no cycle paths possible, TODO: more research about this needed
	optionStmtn := "\tOPTIONS { uniqueEdges: \"path\" }\n"

	vFilterStmnt := ""
	if relation.EntityTo != -1 {
		// If there is a to-node, generate the filter statement
		toConstraints := (*entities)[relation.EntityTo].Constraints
		vFilterStmnt += *createConstraintStatements(&toConstraints, "v", false)

		// Add a WITH statement if the collection of entityTo is not yet included
		if (*entities)[(*relation).EntityFrom].Type != (*entities)[(*relation).EntityTo].Type {
			header = fmt.Sprintf("WITH %v\n %v", (*entities)[(*relation).EntityTo].Type, header)
		}
	}

	relationFilterStmnt := *createConstraintStatements(&relation.Constraints, "p", true)
	footer := "\tLIMIT 1000 \nRETURN DISTINCT p )\n"

	ret := header + forStatement + optionStmtn + vFilterStmnt + relationFilterStmnt + footer
	return &ret
}

func createRelationLetWithOnlyToEntity(relation *relationStruct, name string, entities *[]entityStruct) *string {
	header := fmt.Sprintf("LET %v = (\n\tFOR x IN n%v \n", name, relation.EntityTo)
	forStatement := fmt.Sprintf("\tFOR v, e, p IN %v..%v INBOUND x %s \n", relation.Depth.Min, relation.Depth.Max, relation.Type)

	// Guarantees that there is no path returned with a duplicate edge
	// This way there are no cycle paths possible, TODO: more research about this needed
	optionStmtn := "\tOPTIONS { uniqueEdges: \"path\" }\n"

	relationFilterStmnt := *createConstraintStatements(&relation.Constraints, "p", true)
	footer := "\tLIMIT 1000 \nRETURN DISTINCT p )\n"

	ret := header + forStatement + optionStmtn + relationFilterStmnt + footer
	return &ret
}

// WIP TODO: nodes worden nog niet meegereturned, ff googlen hoe en wat
// func createEntitylessRelationLet(edge *relationStruct, name *string, onlyEdge bool) *string {
// 	var (
// 		header        string
// 		forEdge       string
// 		forSecondNode string
// 	)

// 	footer := "\n\tLIMIT 100\n\tRETURN { vertices: p.vertices[*], edges: p.edges[*] }\n)\n"

// 	// WICKED SWITCHES LETSAGO
// 	if edge.EntityFrom != -1 {
// 		// # 1 (2) Outbound only
// 		header = fmt.Sprintf("LET %v = (\n\tFOR x IN n%v \n", *name, edge.EntityFrom)
// 		forEdge = fmt.Sprintf("\tFOR v, e, p IN %v..%v OUTBOUND x %s \n", edge.Depth.Min, edge.Depth.Max, edge.Type)
// 		if edge.EntityTo != -1 {
// 			// # 2 Node Relation Node
// 			forSecondNode = fmt.Sprintf("\tFILTER v IN n%v \n", edge.EntityTo)
// 		}
// 	} else {
// 		if edge.EntityTo != -1 {
// 			// # 3 Inbound only
// 			header = fmt.Sprintf("LET %v = (\n\tFOR x IN n%v \n", *name, edge.EntityTo)
// 			forEdge = fmt.Sprintf("\tFOR v, e, p IN %v..%v INBOUND x %s \n", edge.Depth.Min, edge.Depth.Max, edge.Type)
// 		} else {
// 			// # 4 Relation with constraints only
// 			header = fmt.Sprintf("LET %v = (\n\tFOR x IN %v \n", *name, edge.Type)
// 			footer = "\tLIMIT 100\n\tRETURN x\n)\n"
// 		}
// 	}
// 	allStatement := true
// 	if onlyEdge {
// 		allStatement = false
// 	}

// 	constraints := *createConstraintStatements(&edge.Constraints, *name, allStatement)

// 	ret := header + forEdge + forSecondNode + constraints + footer
// 	return &ret
// }

/*
#1
{
  "Return": {
    "Entities": [
      0,
      1
    ],
    "Relations": [
      0
    ]
  },
  "Entities": [
    {
      "Type": "airports",
      "Constraints": []
    },
    {
      "Type": "airports",
      "Constraints": []
    }
  ],
  "Relations": [
    {
      "Type": "flights",
      "Depth": {
        "min": 1,
        "max": 1
      },
      "EntityFrom": 0,
      "EntityTo": 1,
      "Constraints": []
    }
  ]
}
*/

/*
#2

{
  "Return": {
    "Entities": [
      0
    ],
    "Relations": [
      0
    ]
  },
  "Entities": [
    {
      "Type": "airports",
      "Constraints": []
    }
  ],
  "Relations": [
    {
      "Type": "flights",
      "Depth": {
        "min": 1,
        "max": 1
      },
      "EntityFrom": 0,
      "EntityTo": -1,
      "Constraints": []
    }
  ]
}
*/

/*
#3
{
  "Return": {
    "Entities": [
      0
    ],
    "Relations": [
      0
    ]
  },
  "Entities": [
    {
      "Type": "airports",
      "Constraints": []
    }
  ],
  "Relations": [
    {
      "Type": "flights",
      "Depth": {
        "min": 1,
        "max": 1
      },
      "EntityFrom": -1,
      "EntityTo": 0,
      "Constraints": []
    }
  ]
}
*/

/*
#4
{
  "Return": {
    "Entities": [],
    "Relations": [
      0
    ]
  },
  "Entities": [],
  "Relations": [
    {
      "Type": "flights",
      "Depth": {
        "min": 1,
        "max": 1
      },
      "EntityFrom": -1,
      "EntityTo": -1,
      "Constraints": [
        {
          "Attribute": "Month",
          "Value": "1",
          "DataType": "number",
          "MatchType": "EQ"
        },
        {
          "Attribute": "Day",
          "Value": "15",
          "DataType": "number",
          "MatchType": "EQ"
        }
      ]
    }
  ]
}
*/

/*
#5
{
  "Return": {
    "Entities": [
      0
    ],
    "Relations": []
  },
  "Entities": [
    {
      "Type": "airports",
      "Constraints": [
        {
          "Attribute": "city",
          "Value": "New York",
          "DataType": "text",
          "MatchType": "exact"
        },
        {
          "Attribute": "country",
          "Value": "USA",
          "DataType": "text",
          "MatchType": "exact"
        }
      ]
    }
  ],
  "Relations": []
}
*/
